import path, { dirname } from "node:path";
import { z } from "zod";
import { MuseError, MuseFileNotFoundError } from "#errors";
import { resolveFilePath } from "#util";
import {
  type MuseEntry,
  type MuseSource,
  loadFromSource,
  parseMuseFile,
} from "./sources";
import { type MusePlugin, loadPlugin } from "./plugins";
import type { UnknownRecord } from "./types";

const SourceSchema = z.union([
  z.object({
    file: z.string(),
  }),
  z.object({
    url: z.string().url(),
  }),
]);

type SourceShorthand = z.infer<typeof SourceSchema>;

// Function to parse an unknown object into parts of a binding
const BindingSchema = z.object({
  contentKey: z.string().default("content"),
  sources: z.array(SourceSchema).default([
    {
      file: "./**/*.{json,yaml,toml,md}",
    },
  ]),
  options: z.record(z.string(), z.any()).default({}),
  processors: z.array(z.string()).default([]),
});

/**
 * The core object of a Muse binding
 * Contains information about the binding file, the configuration of
 * the project, and a list of entries to be processed
 */
export interface Binding<MetaShape = UnknownRecord> {
  /** The raw data read from the binding file */
  readonly _raw: z.infer<typeof BindingSchema>;

  /** Information about the sources used to load entries */
  readonly sources: MuseSource[];

  /** The full file path to the binding file */
  readonly bindingPath: string;

  /** The key used to for default content in unstructured files */
  readonly contentKey: string;

  /** All entries loaded from the binding file */
  readonly entries: MuseEntry[];

  /** A list of processors to be applied to the entries */
  readonly plugins: MusePlugin[];

  /** Arbitrary metadata for processors to use to cache project data */
  readonly meta: MetaShape;

  /** Configuration options for Muse and all processors */
  readonly options: UnknownRecord;
}

function sourceShorthandToSource(shorthand: SourceShorthand): MuseSource {
  if ("file" in shorthand) {
    return {
      location: shorthand.file,
      type: "file",
    };
  }

  if ("url" in shorthand) {
    return {
      location: shorthand.url,
      type: "url",
    };
  }

  throw new MuseError("Invalid source shorthand");
}

/**
 * Given a path, find the binding file
 * If the path is to a directory, check for a binding inside
 * Check order: binding.json, binding.yaml, binding.toml, binding.md
 * Otherwise throw an error
 *
 * @param bindingPath - The path to the binding file
 * @returns The absolute resolved path to the binding file
 */
async function resolveBindingPath(bindingPath: string): Promise<string> {
  const DEFAULT_BINDING_FILES = [
    "binding.json",
    "binding.yaml",
    "binding.toml",
    "binding.md",
  ] as const;
  const inputLocation = Bun.file(bindingPath);

  // If it is a directory, try for the four main supported types
  const stat = await inputLocation.stat();
  if (stat.isDirectory()) {
    for (const bindingFile of DEFAULT_BINDING_FILES) {
      const filePath = path.resolve(bindingPath, bindingFile);
      const resolvedPath = await resolveFilePath(filePath);
      if (resolvedPath) {
        return resolvedPath;
      }
    }

    throw new MuseFileNotFoundError(bindingPath);
  }

  // If not a directory, check if its a file that exists
  const exists = await inputLocation.exists();

  if (!exists) {
    throw new MuseFileNotFoundError(bindingPath);
  }

  // If it is a file, return the path
  return path.resolve(bindingPath);
}

export async function loadBinding(initialPath: string): Promise<Binding> {
  // Resolve the file location if possible
  const bindingPath = await resolveBindingPath(initialPath);
  const bindingDirname = dirname(bindingPath);
  const loadedBindingData = await parseMuseFile(bindingPath, {
    contentKey: "content",
    cwd: bindingDirname,
    ignore: [],
  });

  if (loadedBindingData.length === 0) {
    throw new MuseError("No entries found in binding file");
  }

  // Load and parse the Binding YAML content
  const parsedBinding = BindingSchema.parse(loadedBindingData[0].data);

  const sourceOptions = {
    cwd: bindingDirname,
    contentKey: parsedBinding.contentKey,
    ignore: [bindingPath],
  };

  const sources = parsedBinding.sources.map(sourceShorthandToSource);

  const entries: MuseEntry[] = [];

  for (const source of sources) {
    const entriesFromSource = await loadFromSource(source, sourceOptions);
    entries.push(...entriesFromSource);
  }

  // Load and check plugins
  const plugins: MusePlugin[] = await Promise.all(
    parsedBinding.processors.map(loadPlugin.bind(null, bindingDirname)),
  );

  return {
    _raw: parsedBinding,
    sources,
    bindingPath,
    contentKey: parsedBinding.contentKey,
    entries,
    options: parsedBinding.options,
    plugins,
    meta: {},
  };
}
