import path from "node:path";
import { MuseError, MusePluginError } from "#errors";
import type { Binding } from "./binding";

export type MuseStepFn = (binding: Binding) => Promise<Binding>;

export interface MusePlugin {
  readonly filePath: string;
  readonly name: string;
  readonly description: string;
  readonly version: string;
  readonly step: MuseStepFn;
}

export async function loadPlugin(
  bindingDirname: string,
  pluginPath: string,
): Promise<MusePlugin> {
  // Resolve local paths, use URLs as-is
  const resolvedProcessorPath = pluginPath.startsWith("http")
    ? pluginPath
    : path.resolve(bindingDirname, pluginPath);
  const { step, name, description, version } = await import(
    resolvedProcessorPath
  );

  if (!step || typeof step !== "function") {
    throw new MuseError(
      `Processor at ${pluginPath} does not export a step() function`,
    );
  }

  if (name && typeof name !== "string") {
    throw new MusePluginError(pluginPath, "Plugin name is not a string.");
  }

  if (description && typeof description !== "string") {
    throw new MusePluginError(pluginPath, "Plugin description is not a string");
  }

  if (version && typeof version !== "string") {
    throw new MusePluginError(pluginPath, "Plugin version is not a string");
  }

  return {
    filePath: resolvedProcessorPath,
    step: step as MuseStepFn,
    name: String(name ?? pluginPath),
    description: String(description ?? ""),
    version: String(version ?? "0.0.0"),
  };
}
